// SPDX-FileCopyrightText: Deutsches Elektronen-Synchrotron DESY, MSK, http://msk.desy.de
// SPDX-License-Identifier: LGPL-3.0-or-later

// Define a name for the test module.
#define BOOST_TEST_MODULE testCrcError
// Only after defining the name include the unit test header.
#include "Server.h"
#include "TecDummy.h"

#include <ChimeraTK/ApplicationCore/TestFacility.h>
#include <ChimeraTK/BackendFactory.h>

#include <boost/test/included/unit_test.hpp>

using namespace boost::unit_test_framework;

struct Fixture_t {
  Server theServer{"tec_dummy"};
  ChimeraTK::TestFacility testFacility{theServer};

  boost::shared_ptr<TecDummy> dummy{
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend("ActualDevice"))};
};

/// TestSuite for the server, adapt name
BOOST_AUTO_TEST_SUITE(TecServerCrcError)

/**********************************************************************************************************************/

/// Test serial timeout (as reported by the FPGA) in write operation
BOOST_FIXTURE_TEST_CASE(testCrcErrorWrite, Fixture_t) {
  std::cout << "testCrcErrorWrite" << std::endl;

  auto fanControlEnable = testFacility.getScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_wr");
  auto writeCh1Group7 = testFacility.getVoid("TecModule0/writeCh1Group7_wr");
  auto errorMessage = testFacility.getScalar<std::string>("TecModule0/errorMessage");
  auto errorReset = testFacility.getScalar<int>("TecModule0/errorReset");
  auto errorStatus = testFacility.getScalar<ctk::Boolean>("TecModule0/errorStatus");

  testFacility.runApplication();

  fanControlEnable = 1;
  fanControlEnable.write();
  writeCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 0);
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);

  dummy->simulateCrcError = true;

  fanControlEnable = 42;
  fanControlEnable.write();
  writeCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);
  BOOST_CHECK(!errorMessage.readLatest());
  BOOST_CHECK(!errorStatus.readLatest());
  testFacility.stepApplication();

  // The dummy will have set the value, but answer with a wrong CRC so we should get an error
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 42);
  BOOST_CHECK(errorMessage.readLatest());
  BOOST_CHECK(errorStatus.readLatest());
  BOOST_CHECK(not std::string(errorMessage).empty());
  BOOST_CHECK(errorStatus != 0);

  dummy->simulateCrcError = false;

  writeCh1Group7.write();
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 42);

  // error needs to be reset manually
  BOOST_CHECK(!errorMessage.readLatest());
  BOOST_CHECK(!errorStatus.readLatest());
  errorReset.write();
  testFacility.stepApplication();
  BOOST_CHECK(errorMessage.readLatest());
  BOOST_CHECK(errorStatus.readLatest());
  BOOST_CHECK(std::string(errorMessage).empty());
  BOOST_CHECK(errorStatus == 0);
}

/**********************************************************************************************************************/

BOOST_FIXTURE_TEST_CASE(testSerialTimeoutWriteToFlash, Fixture_t) {
  std::cout << "testCrcErrorWriteToFlash" << std::endl;

  auto fanControlEnable = testFacility.getScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_wr");
  auto writeToFlashCh1Group7 = testFacility.getVoid("TecModule0/writeToFlashCh1Group7_wr");
  auto errorMessage = testFacility.getScalar<std::string>("TecModule0/errorMessage");
  auto errorStatus = testFacility.getScalar<ctk::Boolean>("TecModule0/errorStatus");
  auto errorReset = testFacility.getScalar<int>("TecModule0/errorReset");

  testFacility.runApplication();

  fanControlEnable = 1;
  fanControlEnable.write();
  writeToFlashCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 0);
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);
  BOOST_CHECK(dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanControlEnable));

  dummy->simulateCrcError = true;

  fanControlEnable = 42;
  fanControlEnable.write();
  writeToFlashCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);
  BOOST_CHECK(!errorMessage.readLatest());
  BOOST_CHECK(!errorStatus.readLatest());
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 42);
  BOOST_CHECK(errorMessage.readLatest());
  BOOST_CHECK(errorStatus.readLatest());
  BOOST_CHECK(not std::string(errorMessage).empty());
  BOOST_CHECK(errorStatus);

  dummy->simulateCrcError = false;

  writeToFlashCh1Group7.write();
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 42);
  BOOST_CHECK(dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanControlEnable));

  // error needs to be reset manually
  BOOST_CHECK(!errorMessage.readLatest());
  BOOST_CHECK(!errorStatus.readLatest());
  errorReset.write();
  testFacility.stepApplication();
  BOOST_CHECK(errorMessage.readLatest());
  BOOST_CHECK(errorStatus.readLatest());
  BOOST_CHECK(std::string(errorMessage).empty());
  BOOST_CHECK(not errorStatus);
}

/**********************************************************************************************************************/

BOOST_AUTO_TEST_SUITE_END()
